# -*- coding: utf-8 -*-
##############################################################################
#
#    Author: Nicolas Bessi
#    Copyright 2014
#    Original concept by James Reeves
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License 3
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
from __future__ import unicode_literals
import sys

if sys.version_info < (3, 0): # pragma: no cover
    from urllib import quote
else: # pragma: no cover
    from urllib.parse import quote


def _protect_url(url):
    splited_url = url.split('://', 1)
    if len(splited_url) == 1:
        return quote(url)
    else:
        return "%s://%s" % (splited_url[0], quote(splited_url[1]))


def javascipt_include(*args):
    """Generate elements to include js script

    It generates many `script` tags `<script type.../>

    :param args: sources URLs of script to be included
    :type args: list

    :return: a list representing script tag with URL
    :rtype: list
    """
    tag_list = []
    # we can yield but it is harder to debugg
    for js_url in args:
        tag_list.append(
            ['script',
             {'type': 'text/javascript',
              'src': _protect_url(js_url)}]
        )
    return tag_list


def javascript_with_content(script):
    """Generate element to insert js script

    It generates a `script` tag `<script type...><![CDATA[script]]></script>`

    :param script: script to be interted
    :type script: str

    :return: a list representing script tag with content
    :rtype: list
    """
    tag_list = ['script',
                {'type': 'text/javascript'},
                "<![CDATA[\n %s \n]]>" % script]
    return tag_list


def css_include(*args):
    """Generate elements to include js script

    It generates many `link` tags `<link type="text/css".../>

    :param args:  sources URLs of script to be included
    :type args: list

    :return: a list representing script tag with URL
    :rtype: list
    """
    tag_list = []
    # we can yield but it is harder to debugg
    for css_url in args:
        tag_list.append(
            ['link',
             {'type': 'text/css',
              'href': _protect_url(css_url)}]
        )
    return tag_list


def link_to(url, content):
    """Generate an URL element
    It generates a `a` tag with content `<a href="url">content</a>`

    :param url: destination URL
    :type url: str

    :return: a list representing `a` link tag
    :rtype: list
    """
    tag_list = ['a',
                {'href': _protect_url(url)},
                content]
    return tag_list


def image(source_url, alt=None, height=None, width=None):
    """Generate an image element
    It generates a `image` from URL <img src="url" alt=.../>

    :param url: image source URL
    :type url: str

    :param alt: alternate text
    :type alt: str

    :param height: height of image
    :param height: str or numeric

    :param width: width of image
    :param width: str or numeric

    :return: a list representing an `img` tag
    :rtype: list
    """
    attrs = {'src': _protect_url(source_url)}
    if alt:
        attrs['alt'] = alt
    if height:
        attrs['height'] = height
    if width:
        attrs['width'] = width
    tag_list = ['img', attrs]
    return tag_list



# -*- coding: utf-8 -*-
##############################################################################
#
#    Author: Nicolas Bessi
#    Copyright 2014
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License 3
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
import re

try:
    import unittest2 as unittest
except ImportError:
    import unittest

CLEAN_REGEX = re.compile('[\t\r\n]')


class CommonTest(unittest.TestCase):

    def normalize_result(self, raw_string):
        """Remove tab, carriage and space around balise

        :param raw_string: HTML/XML to clean
        :type raw_string: basestring

        :return: normalized string
        :rtype: str

        """
        return CLEAN_REGEX.sub('', raw_string)

    def test_normalize_result(self):
        """Test normalize helper"""
        raw = u"Hello\t \nWorld"
        clean = self.normalize_result(raw)
        self.assertEqual(clean, u"Hello World")



# -*- coding: utf-8 -*-
##############################################################################
#
#    Author: Nicolas Bessi
#    Copyright 2014
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License 3
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
from __future__ import unicode_literals
import logging

from ..core import html, xml, _logger, convert
from .common import CommonTest


class HTMLTest(CommonTest):

    def test_comprehension_conversion(self):
        """Test basic HTML and list comprehension"""
        data = [
            [u'div',
             {u'class': 'a-class', 'data-y': '23'},
             [u'span', 'my-text',
              [u'ul', [['li', x] for x in [u'café', u'milk', u'sugar']]]]]
        ]

        awaited = ('<!DOCTYPE html><html dir="rtl" lang="en" xml:lang="en">'
                   '<div class="a-class" data-y="23"><span>my-text'
                   '<ul><li>caf\xe9</li><li>milk</li><li>sugar</li></ul>'
                   '</span></div></html>')
        conv = html(data)
        self.assertEquals(awaited, self.normalize_result(conv))

    def test_html5_doc_type(self):
        """Test HTML 5 DOCTYPE"""
        data = []
        awaited = '<!DOCTYPE html><html dir="rtl" lang="en" xml:lang="en"/>'
        conv = html(data, etype='html5')
        self.assertEquals(awaited, self.normalize_result(conv))

    def test_html4_doc_type(self):
        """Test HTML 4 DOCTYPE"""
        data = []
        awaited = ('<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN" '
                   '"http://www.w3.org/TR/html4/strict.dtd"><html dir="rtl" lang="en"'
                   ' xml:lang="en"/>')
        conv = html(data, etype='html4')
        self.assertEquals(awaited, self.normalize_result(conv))

    def test_xhtml_strict_doc_type(self):
        """Test XHTML strict DOCTYPE"""
        data = []
        awaited = ('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" '
                   '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">'
                   '<html dir="rtl" lang="en" xml:lang="en" '
                   'xmlns="http://www.w3.org/1999/xhtml"/>')
        conv = html(data, etype='xhtml-strict')
        self.assertEquals(awaited, self.normalize_result(conv))

    def test_xhtml_transitional_doc_type(self):
        """Test XHTML transitional DOCTYPE"""
        data = []
        awaited = ('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML '
                   '1.0 Transitional//EN" '
                   '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'
                   '<html dir="rtl" lang="en" xml:lang="en" '
                   'xmlns="http://www.w3.org/1999/xhtml"/>')
        conv = html(data, etype='xhtml-transitional')
        self.assertEquals(awaited, self.normalize_result(conv))

    def test_convert(self):
        """Test block_convert"""
        data = ['div']
        awaited = '<div/>'
        conv = convert(data)
        self.assertEquals(awaited, self.normalize_result(conv))

    def test_wrong_type(self):
        data = []
        self.assertRaises(ValueError, html, data, 'kaboom')

    def test_debug_logging(self):
        """Ensure that debug logging will not break"""
        current_level = _logger.getEffectiveLevel()
        _logger.setLevel(logging.DEBUG)
        html([])
        _logger.setLevel(current_level)


class XMLTest(CommonTest):

    def test_minimal_xml(self):
        """Test minimal XML"""
        data = ['form-desc',
                ['field', {'name': 'a_name'}],
                ['field', {'name': 'a_other_name'}]]
        conv = xml(data, 'foo-ns', bar='an_attr')
        awaited = ('<?xml version="1.0" encoding="UTF-8"?>'
                   '<foo-ns bar="an_attr"><form-desc><field name="a_name"/>'
                   '<field name="a_other_name"/></form-desc></foo-ns>')
        self.assertEquals(awaited, self.normalize_result(conv))



# -*- coding: utf-8 -*-
##############################################################################
#
#    Author: Nicolas Bessi
#    Copyright 2014
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License 3
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
from __future__ import unicode_literals
from ..element import (link_to, javascript_with_content, javascipt_include,
                       css_include, image)
from ..core import _convert_tree
from .common import CommonTest


class ElementTest(CommonTest):

    def test_javascript_include(self):
        """Test js inclusion helper"""
        data = javascipt_include("jquery-1.11.1.min.js")
        awaited = '<script src="jquery-1.11.1.min.js" type="text/javascript"/>'
        conv = ''.join(_convert_tree(data))
        self.assertEquals(awaited, self.normalize_result(conv))

    def test_javascript_include_remote(self):
        """Test js inclusion helper using remote URL"""
        url = "http://an-url/jquery.mobile.min.css"
        data = javascipt_include(url)
        awaited = ('<script src="http://an-url/jquery.mobile.min.css" type="text/javascript"/>')
        conv = ''.join(_convert_tree(data))
        self.assertEquals(awaited, self.normalize_result(conv))

    def test_javascript_include_mutlipe(self):
        """Test js include helper using many URL"""
        url = "http://an-url/jquery.mobile.min.css"
        data = javascipt_include(url, "jquery-1.11.1.min.js")
        awaited = ('<script src="http://an-url/jquery.mobile.min.css" type="text/javascript"/>'
                   '<script src="jquery-1.11.1.min.js" type="text/javascript"/>')
        conv = ''.join(_convert_tree(data))
        self.assertEquals(awaited, self.normalize_result(conv))

    def test_javascript_with_content(self):
        """Test javascript with content helper"""
        data = javascript_with_content("alert('Hello, World!')")
        awaited = ("<script type=\"text/javascript\">"
                   "<![CDATA[ alert('Hello, World!') ]]>"
                   "</script>")
        conv = ''.join(_convert_tree(data))
        self.assertEquals(awaited, self.normalize_result(conv))

    def test_link_to(self):
        """Test URL helper"""
        data = link_to('http://github.com/~pyhiccup', 'pyhiccup')
        awaited = '<a href="http://github.com/%7Epyhiccup">pyhiccup</a>'
        conv = ''.join(_convert_tree(data))
        self.assertEquals(awaited, self.normalize_result(conv))

    def test_css_include_multiple(self):
        """Test CSS helper"""
        data = css_include('http://homer.com/css', 'local.css')
        awaited = ('<link href="http://homer.com/css" type="text/css"/>'
                   '<link href="local.css" type="text/css"/>')

        conv = ''.join(_convert_tree(data))
        self.assertEquals(awaited, self.normalize_result(conv))

    def test_image(self):
        """Test image helper"""
        data = image('http://homer.com/tho.jpg')
        awaited = ('<img src="http://homer.com/tho.jpg"/>')
        conv = ''.join(_convert_tree(data))
        self.assertEquals(awaited, self.normalize_result(conv))

    def test_image_with_aprams(self):
        """Test image helper"""
        data = image('http://homer.com/tho.jpg',
                     alt='donut',
                     height='20px',
                     width='10px')
        awaited = ('<img alt="donut" height="20px" '
                   'src="http://homer.com/tho.jpg" '
                   'width="10px"/>')
        conv = ''.join(_convert_tree(data))
        self.assertEquals(awaited, self.normalize_result(conv))



# -*- coding: utf-8 -*-
##############################################################################
#
#    Author: Nicolas Bessi
#    Copyright 2014
#    Original concept by James Reeves
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License 3
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
from __future__ import unicode_literals

DOC_TYPES = {
    'html4': "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01//EN\" "
             "\"http://www.w3.org/TR/html4/strict.dtd\">\n",

    'xhtml-strict': "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 ""Strict//EN\" "
                    "\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n",

    'xhtml-transitional': "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" "
                          "\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n",

    'html5': "<!DOCTYPE html>\n",
}

DEFAULT_XMLNS = 'http://www.w3.org/1999/xhtml'
XMl_DECLARATION = '<?xml version="1.0" encoding="UTF-8"?>'


def get_doc_type(doc_type):
    """Return a DOCTYPE declaration

    :param doc_type: doc type string must be in ``page.DOC_TYPES``
    :type doc_type: str
    :return: DOCTYPE declaration
    :rtype: str

    """
    if doc_type not in DOC_TYPES:
        raise ValueError(
            'Invalid DOCTYPE %s available values are %s' %
            (doc_type, DOC_TYPES.keys())
        )
    return DOC_TYPES[doc_type]


def build_html_enclosing_tag(etype, **kwargs):
    """Generate html tag list representation

    :param etype: html doc type `html5, html4, xhtml-strict,
                  xhtml-transitional`
    :type etype: str
    :param kwargs: dict of attribute for HTML tag will override defaults
    :type kwargs: dict

    :return: html tag list representation ['html', {'xmlns': ...}]
    :rtype: dict
    """
    attrs = {}
    if etype in DOC_TYPES:
        attrs['lang'] = 'en'
        attrs['dir'] = 'rtl'
        attrs['xml:lang'] = 'en'
    if 'xhtml' in etype:
        attrs[u'xmlns'] = DEFAULT_XMLNS
    attrs.update(kwargs)
    return ['html', attrs]


def build_xml_enclosing_tag(etype, **kwargs):
    """Generate XML root tag list representation

    :param etype: root tag name
    :type etype: str
    :param kwargs: dict of attribute for root tag
    :type kwargs: dict

    :return: root xml tag list representation ['atag', {'attr': ...}]
    :rtype: dict
    """
    return [etype, kwargs]



# -*- coding: utf-8 -*-
##############################################################################
#
#    Author: Nicolas Bessi
#    Copyright 2014
#    Original concept by James Reeves
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License 3
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
from __future__ import unicode_literals
import logging
import copy
from itertools import chain

from .page import get_doc_type
from .page import build_html_enclosing_tag, build_xml_enclosing_tag
from .page import XMl_DECLARATION

_logger = logging.getLogger('pyhiccup.convert')

TREE_TYPE = (list, tuple)


def format_attributes(attributes):
    """Transform dict to XMl/HTMl attributes

    {'class': 'a-class', 'data-sel': 'a-sel'}
    returns ``class="a-class" data-sel="a-sel"``

    :param attributes: dict of XML/HMTL attributes to transform
                       to string
    :type attributes: dict

    :return: a list of attributes eg. ``class="a-class" data-sel="a-sel"``
    :rtype: str
    """
    output = []
    for item in sorted(attributes.items()):
        output.append('%s=\"%s\"' % item)
    return " %s" % ' '.join(output)


def _convert_tree(node):
    """Transform a list describing HTML leaf to a list of HTML string

    ready to be joined

    :param args: leaf list describing HTML
    :type args: list, tuple

    :return: a list of string
    :rtype: list
    """
    #perfo tweak with side effect
    if isinstance(node[0], TREE_TYPE):
        for sub_node in node:
            for x in _convert_tree(sub_node):
                yield x
        return
    btype = node[0]
    rest = node[1:] if len(node) > 1 else []
    attrs = ''
    inner_trees = []
    inner_element = ''
    for element in rest:
        if not element:
            continue
        if isinstance(element, TREE_TYPE):
            if isinstance(element[0], TREE_TYPE):
                inner_trees.extend(element)
            else:
                inner_trees.append(element)
        elif isinstance(element, dict):
            attrs = format_attributes(element)
        else:
            inner_element = element
    if inner_element or inner_trees:
        yield '<%s%s>' % (
            btype,
            attrs,
        )
        yield inner_element
        if inner_trees:
            for ext in inner_trees:
                for x in _convert_tree(ext):
                    yield x
        yield '</%s>' % btype
    else:
        yield '<%s%s/>' % (
            btype,
            attrs,
        )


def _inclose_page(declaration, enclosing_tag, value):
    """Take a page code and a value
    and inclose value into page code
    :param declaration: Declaration of the document
    :type declaration: str
    :param enclosing_tag: page code list type e.g. `['html', {'xmlns':...}]`
    :type enclosing_tag: str

    :param value: the list to be converted to *ML
    :type value: str
    """
    to_convert = copy.deepcopy(enclosing_tag)
    to_convert.append(value)
    converted = chain(
        [declaration],
        _convert_tree(to_convert)
    )
    if _logger.getEffectiveLevel() == logging.DEBUG:
        _logger.debug(
            list(chain([declaration],
                       _convert_tree(to_convert)))
        )
    return converted


def html(value, etype='html5', **kwargs):
    """Transform a list describing HTML to raw HTML

    :param value: list of list describing HTML
    :type value: list, tuple
    :param etype: enclosing type `html5, html4, xhtml-strict,
                  xhtml-transitional`
    :type etype: str
    :param kwargs: dict of enclosing tag attributes
    :type kwargs: dict

    :return: HTML string representation
    :rtype: str, unicode
    """
    declaration = get_doc_type(etype)
    enclosing_tag = build_html_enclosing_tag(etype)
    converted = _inclose_page(declaration, enclosing_tag, value)
    return ''.join(converted)


def xml(value, etype, **kwargs):
    """Transform a list describing XML to raw XML

    :param value: list of list describing XML
    :type value: list, tuple
    :param etype: enclosing type the root tag name
    :type etype: str
    :param kwargs: dict of enclosing tag attributes
    :type kwargs: dict

    :return: XML string representation
    :rtype: str, unicode
    """
    declaration = XMl_DECLARATION
    enclosing_tag = build_xml_enclosing_tag(etype, **kwargs)
    converted = _inclose_page(declaration, enclosing_tag, value)
    return ''.join(converted)

def convert(value):
    """Transform a list to arbitratry XML

    :param value: list of list describing XML
    :type value: list, tuple

    :return: XML string representation
    :rtype: str, unicode
    """
    return ''.join(_convert_tree(value))
